@@ -18,30 +18,26 @@ import bisect
 import sys
 import token
 from ast import Module
-from typing import Iterable, Iterator, List, Optional, Tuple, Any, cast, TYPE_CHECKING, Type
-
-import six
-from six.moves import xrange  # pylint: disable=redefined-builtin
+from typing import Iterable, Iterator, List, Optional, Tuple, Any, cast, TYPE_CHECKING
 
 from .line_numbers import LineNumbers
-from .util import Token, match_token, is_non_coding_token, patched_generate_tokens, last_stmt, annotate_fstring_nodes, generate_tokens
+from .util import (
+  Token, match_token, is_non_coding_token, patched_generate_tokens, last_stmt,
+  annotate_fstring_nodes, generate_tokens, is_module, is_stmt
+)
 
 if TYPE_CHECKING:  # pragma: no cover
   from .util import AstNode, TokenInfo
 
 
-class ASTTextBase(six.with_metaclass(abc.ABCMeta, object)):
-  def __init__(self, source_text, filename):
-    # type: (Any, str) -> None
-    # FIXME: Strictly, the type of source_text is one of the six string types, but hard to specify with mypy given
-    # https://mypy.readthedocs.io/en/stable/common_issues.html#variables-vs-type-aliases
-
+class ASTTextBase(metaclass=abc.ABCMeta):
+  def __init__(self, source_text: str, filename: str) -> None:
     self._filename = filename
 
     # Decode source after parsing to let Python 2 handle coding declarations.
     # (If the encoding was not utf-8 compatible, then even if it parses correctly,
     # we'll fail with a unicode error here.)
-    source_text = six.ensure_text(source_text)
+    source_text = str(source_text)
 
     self._text = source_text
     self._line_numbers = LineNumbers(source_text)
@@ -58,7 +54,7 @@ class ASTTextBase(six.with_metaclass(abc.ABCMeta, object)):
     This means that if ``padded`` is True, the start position will be adjusted to include
     leading whitespace if ``node`` is a multiline statement.
     """
-    raise NotImplementedError
+    raise NotImplementedError  # pragma: no cover
 
   def get_text_range(self, node, padded=True):
     # type: (AstNode, bool) -> Tuple[int, int]
@@ -86,7 +82,7 @@ class ASTTextBase(six.with_metaclass(abc.ABCMeta, object)):
     return self._text[start: end]
 
 
-class ASTTokens(ASTTextBase, object):
+class ASTTokens(ASTTextBase):
   """
   ASTTokens maintains the text of Python code in several forms: as a string, as line numbers, and
   as tokens, and is used to mark and access token and position information.
@@ -108,9 +104,6 @@ class ASTTokens(ASTTextBase, object):
 
   def __init__(self, source_text, parse=False, tree=None, filename='<unknown>', tokens=None):
     # type: (Any, bool, Optional[Module], str, Iterable[TokenInfo]) -> None
-    # FIXME: Strictly, the type of source_text is one of the six string types, but hard to specify with mypy given
-    # https://mypy.readthedocs.io/en/stable/common_issues.html#variables-vs-type-aliases
-
     super(ASTTokens, self).__init__(source_text, filename)
 
     self._tree = ast.parse(source_text, filename) if parse else tree
@@ -131,11 +124,11 @@ class ASTTokens(ASTTextBase, object):
     """
     Given the root of the AST or Astroid tree produced from source_text, visits all nodes marking
     them with token and position information by adding ``.first_token`` and
-    ``.last_token``attributes. This is done automatically in the constructor when ``parse`` or
+    ``.last_token`` attributes. This is done automatically in the constructor when ``parse`` or
     ``tree`` arguments are set, but may be used manually with a separate AST or Astroid tree.
     """
     # The hard work of this class is done by MarkTokens
-    from .mark_tokens import MarkTokens # to avoid import loops
+    from .mark_tokens import MarkTokens  # to avoid import loops
     MarkTokens(self).visit_tree(root_node)
 
   def _translate_tokens(self, original_tokens):
@@ -228,7 +221,7 @@ class ASTTokens(ASTTextBase, object):
     """
     Looks for the first token, starting at start_token, that matches tok_type and, if given, the
     token string. Searches backwards if reverse is True. Returns ENDMARKER token if not found (you
-    can check it with `token.ISEOF(t.type)`.
+    can check it with `token.ISEOF(t.type)`).
     """
     t = start_token
     advance = self.prev_token if reverse else self.next_token
@@ -246,7 +239,7 @@ class ASTTokens(ASTTextBase, object):
     Yields all tokens in order from first_token through and including last_token. If
     include_extra is True, includes non-coding tokens such as tokenize.NL and .COMMENT.
     """
-    for i in xrange(first_token.index, last_token.index + 1):
+    for i in range(first_token.index, last_token.index + 1):
       if include_extra or not is_non_coding_token(self._tokens[i].type):
         yield self._tokens[i]
 
@@ -281,7 +274,7 @@ class ASTTokens(ASTTextBase, object):
     return start, end
 
 
-class ASTText(ASTTextBase, object):
+class ASTText(ASTTextBase):
   """
   Supports the same ``get_text*`` methods as ``ASTTokens``,
   but uses the AST to determine the text positions instead of tokens.
@@ -289,8 +282,6 @@ class ASTText(ASTTextBase, object):
 
   It also (sometimes) supports nodes inside f-strings, which ``ASTTokens`` doesn't.
 
-  Astroid trees are not supported at all and will raise an error.
-
   Some node types and/or Python versions are not supported.
   In these cases the ``get_text*`` methods will fall back to using ``ASTTokens``
   which incurs the usual setup cost the first time.
@@ -298,12 +289,6 @@ class ASTText(ASTTextBase, object):
   """
   def __init__(self, source_text, tree=None, filename='<unknown>'):
     # type: (Any, Optional[Module], str) -> None
-    # FIXME: Strictly, the type of source_text is one of the six string types, but hard to specify with mypy given
-    # https://mypy.readthedocs.io/en/stable/common_issues.html#variables-vs-type-aliases
-
-    if not isinstance(tree, (ast.AST, type(None))):
-      raise NotImplementedError('ASTText only supports AST trees')
-
     super(ASTText, self).__init__(source_text, filename)
 
     self._tree = tree
@@ -332,26 +317,27 @@ class ASTText(ASTTextBase, object):
     return self._asttokens
 
   def _get_text_positions_tokenless(self, node, padded):
-    # type: (ast.AST, bool) -> Tuple[Tuple[int, int], Tuple[int, int]]
+    # type: (AstNode, bool) -> Tuple[Tuple[int, int], Tuple[int, int]]
     """
     Version of ``get_text_positions()`` that doesn't use tokens.
     """
-    if sys.version_info[:2] < (3, 8):
-      raise AssertionError("This method should only be called internally after checking supports_tokenless()")
-
-    if isinstance(node, ast.Module):
+    if is_module(node):
       # Modules don't have position info, so just return the range of the whole text.
       # The token-using method does something different, but its behavior seems weird and inconsistent.
       # For example, in a file with only comments, it only returns the first line.
       # It's hard to imagine a case when this matters.
       return (1, 0), self._line_numbers.offset_to_line(len(self._text))
 
-    if not hasattr(node, 'lineno'):
+    if getattr(node, 'lineno', None) is None:
       return (1, 0), (1, 0)
 
     assert node  # tell mypy that node is not None, which we allowed up to here for compatibility
 
     decorators = getattr(node, 'decorator_list', [])
+    if not decorators:
+      # Astroid uses node.decorators.nodes instead of node.decorator_list.
+      decorators_node = getattr(node, 'decorators', None)
+      decorators = getattr(decorators_node, 'nodes', [])
     if decorators:
       # Function/Class definition nodes are marked by AST as starting at def/class,
       # not the first decorator. This doesn't match the token-using behavior,
@@ -360,18 +346,35 @@ class ASTText(ASTTextBase, object):
     else:
       start_node = node
 
-    if padded and last_stmt(node).lineno != node.lineno:
-      # Include leading indentation for multiline statements.
+    start_lineno = start_node.lineno
+    end_node = last_stmt(node)
+
+    # Include leading indentation for multiline statements.
+    # This doesn't mean simple statements that happen to be on multiple lines,
+    # but compound statements where inner indentation matters.
+    # So we don't just compare node.lineno and node.end_lineno,
+    # we check for a contained statement starting on a different line.
+    if padded and (
+        start_lineno != end_node.lineno
+        or (
+            # Astroid docstrings aren't treated as separate statements.
+            # So to handle function/class definitions with a docstring but no other body,
+            # we just check that the node is a statement with a docstring
+            # and spanning multiple lines in the simple, literal sense.
+            start_lineno != node.end_lineno
+            and getattr(node, "doc_node", None)
+            and is_stmt(node)
+        )
+    ):
       start_col_offset = 0
     else:
-      start_col_offset = self._line_numbers.from_utf8_col(start_node.lineno, start_node.col_offset)
+      start_col_offset = self._line_numbers.from_utf8_col(start_lineno, start_node.col_offset)
 
-    start = (start_node.lineno, start_col_offset)
+    start = (start_lineno, start_col_offset)
 
     # To match the token-using behaviour, we exclude trailing semicolons and comments.
     # This means that for blocks containing multiple statements, we have to use the last one
     # instead of the actual node for end_lineno and end_col_offset.
-    end_node = last_stmt(node)
     end_lineno = cast(int, end_node.end_lineno)
     end_col_offset = cast(int, end_node.end_col_offset)
     end_col_offset = self._line_numbers.from_utf8_col(end_lineno, end_col_offset)
@@ -400,20 +403,14 @@ class ASTText(ASTTextBase, object):
     return self.asttokens.get_text_positions(node, padded)
 
 
-# Node types that _get_text_positions_tokenless doesn't support. Only relevant for Python 3.8+.
-_unsupported_tokenless_types = ()  # type: Tuple[Type[ast.AST], ...]
-if sys.version_info[:2] >= (3, 8):
-  _unsupported_tokenless_types += (
-    # no lineno
-    ast.arguments, ast.withitem,
-  )
-  if sys.version_info[:2] == (3, 8):
-    _unsupported_tokenless_types += (
-      # _get_text_positions_tokenless works incorrectly for these types due to bugs in Python 3.8.
-      ast.arg, ast.Starred,
-      # no lineno in 3.8
-      ast.Slice, ast.ExtSlice, ast.Index, ast.keyword,
-    )
+# Node types that _get_text_positions_tokenless doesn't support.
+# These initial values are missing lineno.
+_unsupported_tokenless_types = ("arguments", "Arguments", "withitem")  # type: Tuple[str, ...]
+if sys.version_info[:2] == (3, 8):
+  # _get_text_positions_tokenless works incorrectly for these types due to bugs in Python 3.8.
+  _unsupported_tokenless_types += ("arg", "Starred")
+  # no lineno in 3.8
+  _unsupported_tokenless_types += ("Slice", "ExtSlice", "Index", "keyword")
 
 
 def supports_tokenless(node=None):
@@ -425,10 +422,11 @@ def supports_tokenless(node=None):
 
   The following cases are not supported:
 
-    - Python 3.7 and earlier
     - PyPy
-    - Astroid nodes (``get_text*`` methods of ``ASTText`` will raise an error)
-    - ``ast.arguments`` and ``ast.withitem``
+    - ``ast.arguments`` / ``astroid.Arguments``
+    - ``ast.withitem``
+    - ``astroid.Comprehension``
+    - ``astroid.AssignName`` inside ``astroid.Arguments`` or ``astroid.ExceptHandler``
     - The following nodes in Python 3.8 only:
       - ``ast.arg``
       - ``ast.Starred``
@@ -438,8 +436,15 @@ def supports_tokenless(node=None):
       - ``ast.keyword``
   """
   return (
-      isinstance(node, (ast.AST, type(None)))
-      and not isinstance(node, _unsupported_tokenless_types)
-      and sys.version_info[:2] >= (3, 8)
+      type(node).__name__ not in _unsupported_tokenless_types
+      and not (
+        # astroid nodes
+        not isinstance(node, ast.AST) and node is not None and (
+          (
+            type(node).__name__ == "AssignName"
+            and type(node.parent).__name__ in ("Arguments", "ExceptHandler")
+          )
+        )
+      )
       and 'pypy' not in sys.version.lower()
   )
